//========================================================================
// tb_RegIncrNstage
//========================================================================
// A Verilog test bench for our parametrized N-stage RegIncr unit

`default_nettype none
`timescale 1ps/1ps

`include "regincr/RegIncrNstage.v"
`include "vc/TestRandDelaySource.v"
`include "vc/TestRandDelaySink.v"
`include "vc/trace.v"

//------------------------------------------------------------------------
// Testbench defines
//------------------------------------------------------------------------

localparam NUM_TESTS = 3;
localparam NUM_STAGES = 5;

localparam  INPUT_TEST_SIZE = 8;
localparam OUTPUT_TEST_SIZE = 8;

//------------------------------------------------------------------------
// Top-level module
//------------------------------------------------------------------------

module top(  input logic clk, input logic linetrace );
  
  // DUT signals
  logic        reset;

  logic                        istream_val;
  logic [ INPUT_TEST_SIZE-1:0] istream_msg;

  logic                        ostream_val;
  logic [OUTPUT_TEST_SIZE-1:0] ostream_msg;

  // Source and sink messages

  logic [  INPUT_TEST_SIZE-1:0 ] src_msgs [ NUM_TESTS-1:0 ];
  logic [ OUTPUT_TEST_SIZE-1:0 ] snk_msgs [ NUM_TESTS-1:0 ];

  // Signals to indicate completion

  logic src_done;
  logic snk_done;

  //----------------------------------------------------------------------
  // Module instantiations
  //----------------------------------------------------------------------

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // Test source
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  vc_TestSource 
  #(
    .p_msg_nbits ( INPUT_TEST_SIZE ),
    .p_num_msgs  (       NUM_TESTS )
  ) src (
    .clk         (             clk ),
    .reset       (           reset ),

    .val         (     istream_val ),
    .rdy         (            1'b1 ), // Always rdy
    .msg         (     istream_msg ),

    .done        (        src_done )
  );

  assign src.m = src_msgs;
  
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // DUT
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  
  TUT4_VERILOG_regincr_RegIncrNstage #( NUM_STAGES ) dut
  (
    .clk         (       clk   ),
    .reset       (       reset ),

    .in_         ( istream_msg ),

    .out         ( ostream_msg )
  );

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // Test sink
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  vc_TestSink
  #(
    .p_msg_nbits ( OUTPUT_TEST_SIZE ),
    .p_num_msgs  (        NUM_TESTS ),
    .p_sim_mode  (                1 )
  ) sink (
    .clk         (              clk ),
    .reset       (            reset ),

    .val         (      ostream_val ),
    /* verilator lint_off PINCONNECTEMPTY */
    .rdy         (                  ), // Unused
    /* verilator lint_off PINCONNECTEMPTY */
    .msg         (      ostream_msg ),

    .done        (         snk_done )
  );

  assign sink.m = snk_msgs;

  //----------------------------------------------------------------------
  // Keep track of message validity
  //----------------------------------------------------------------------

  logic val_arr [ NUM_STAGES:0 ] ;
  assign val_arr[0]  = istream_val;
  assign ostream_val = val_arr[ NUM_STAGES ];

  genvar i;

  generate
    for( i = 0; i < NUM_STAGES; i++ ) begin
      always @( posedge clk ) begin
        val_arr[ i + 1 ] <= val_arr[ i ];
      end
    end
  endgenerate

  //----------------------------------------------------------------------
  // Task for adding test cases
  //----------------------------------------------------------------------

  task test_case(
    input logic [  INPUT_TEST_SIZE-1:0 ] src_msg,
    input logic [ OUTPUT_TEST_SIZE-1:0 ] snk_msg
  );
  begin
    integer idx = 0;

    // Add messages to test arrays
    src_msgs[ idx ] = src_msg;
    snk_msgs[ idx ] = snk_msg;

    idx = idx + 1;
  end
  endtask

  //----------------------------------------------------------------------
  // Test cases
  //----------------------------------------------------------------------
  // Don't forget to change NUM_TESTS above when adding new tests!

  initial begin

    // Test cases

    test_case( 8'd1, 8'd1 + NUM_STAGES );
    test_case( 8'd4, 8'd4 + NUM_STAGES );
    test_case( 8'd9, 8'd9 + NUM_STAGES );

  end

  //----------------------------------------------------------------------
  // Run the Test Bench
  //----------------------------------------------------------------------

  initial begin

    $display( "Starting tb_RegIncrNstage..." );
    reset = 1;
    
    // Wait a bit, then de-assert reset on negedge
    #10 
    @( negedge clk );
    reset = 0;

    // Wait for the test to finish
    while( !snk_done ) @( negedge clk );

    // Check that the source is also done
    if( !src_done )
      $error( "[ FAILED ] Our sink received more messages than our source has!" );
    else
      $display( "The testbench has finished" );

    // Delay for a bit for a better waveform
    #100
    $finish;
  end

  //----------------------------------------------------------------------
  // Timeout
  //----------------------------------------------------------------------
  // This is to ensure that our test eventually finishes, even if the sink
  // isn't receiving messages

  initial begin
    for( integer j = 0; j < 1000000; j = j + 1 ) begin
      @( negedge clk );
    end

    $error( "TIMEOUT: Testbench didn't finish in time" );
    $finish;
  end

endmodule
